@@ -1,43 +1,55 @@
-#' Enable LLM to call R functions
+# `answer_using_tools()` prompt wrap --------------------------------------
+
+#' @title
+#' Enable LLM to call R functions (and/or MCP server tools)
 #'
+#' @description
 #' This function adds the ability for the a LLM to call R functions.
 #' Users can specify a list of functions that the LLM can call, and the
 #' prompt will be modified to include information, as well as an
 #' accompanying extraction function to call the functions (handled by
 #' [send_prompt()]). Documentation for the functions is extracted from
 #' the help file (if available), or from documentation added by
-#' [tools_add_docs()]
+#' [tools_add_docs()]. Users can also provide an 'ellmer' tool definition
+#' (see [ellmer::tool()]; ['ellmer' documentation](https://ellmer.tidyverse.org/articles/tool-calling.html)).
+#' Model Context Protocol (MCP) tools from MCP servers, as returned from
+#' [mcptools::mcp_tools()], may also be used. Regardless of which type of
+#' tool definition is provided, the function will work with both 'ellmer' and
+#' regular LLM providers (the function converts between the two types as needed).
 #'
-#' @details Note that this method of function calling is purely text-based.
-#' This makes it suitable for any LLM and any LLM provider. However,
-#' 'native' function calling (where the LLM model provider restricts the
-#' model to special tokens that can be used to call functions) may perform
-#' better in terms of accuracy and efficiency. 'tidyprompt' may support
-#' 'native' function calling in the future
+#' @details
+#' Note that conversion between 'tidyprompt' and 'ellmer' tool definitions
+#' is experimntal and might contain bugs.
 #'
 #' @param prompt A single string or a [tidyprompt()] object
 #'
-#' @param tools An R function or a list of R functions that the LLM can call.
-#' If the function has been documented in a help file (e.g., because it is part of a
-#' package), the documentation will be parsed from the help file. If it is a custom
-#' function, documentation should be added with [tools_add_docs()]
+#' @param tools An R function, an 'ellmer' tool definition (from [ellmer::tool()]),
+#' or a list of either, which the LLM will be able to call. If an R function is passed
+#' which has been documented in a help file (e.g., because it is part of a package),
+#' the documentation will be parsed from the help file. If it is a custom function,
+#' documentation should be added with [tools_add_docs()] or with [ellmer::tool()].
+#' Note that you can also provide Model Context Protocol (MCP) tools from MCP servers
+#' as returned from [mcptools::mcp_tools()]
 #'
 #' @param type (optional) The way that tool calling should be enabled.
-#' 'auto' will automatically determine the type based on 'llm_provider$api_type'
-#' (note that this does not consider model compatibility, and could lead to errors;
-#' set 'type' manually if errors occur). 'openai' and 'ollama' will set the
-#' relevant API parameters. 'text-based' will provide function definitions
-#' in the prompt, extract function calls from the LLM response, and call the
-#' functions, providing the results back via [llm_feedback()]. 'text-based'
+#' "auto" will automatically determine the type based on `llm_provider$api_type`
+#' or 'llm_provider$tool_type' (if set; 'tool_type' overrides 'api_type' determination)
+#' (note that this may not consider model compatibility, and could lead to errors;
+#' set 'type' manually if errors occur). "openai" and "ollama" will set the
+#' relevant API parameters. "ellmer" will register the tool in the 'ellmer' chat
+#' object of the LLM provider; note that this will only work for an [llm_provider_ellmer()]
+#' ("auto" will always set the type to "ellmer" if you are using an 'ellmer' LLM provider).
+#' "text-based" will provide function definitions in the prompt, extract function
+#' calls from the LLM response, and call the
+#' functions, providing the results back via [llm_feedback()]. "text-based"
 #' always works, but may be inefficient for APIs that support tool calling
-#' natively. However, 'text-based' may be more reliable and flexible, especially
-#' when combining with other prompt wraps. 'openai' and 'ollama' may not allow
-#' for retries if the function call did not provide the expected result. Note that
-#' when using 'openai' or 'ollama', tool calls are not counted as interactions and
-#' may continue indefinitely (use with caution)
+#' natively. Note that when using "openai", "ollama", or "ellmer", tool calls are not counted
+#' as interactions by [send_prompt()] and may continue indefinitely unless restricted
+#' by other means
 #'
 #' @return A [tidyprompt()] with an added [prompt_wrap()] which
-#' will allow the LLM to call R functions
+#' will allow the LLM to call the given R functions when evaluating
+#' the prompt with [send_prompt()]
 #'
 #' @export
 #'
@@ -51,7 +63,7 @@
 answer_using_tools <- function(
   prompt,
   tools = list(),
-  type = c("text-based", "auto", "openai", "ollama")
+  type = c("auto", "openai", "ollama", "ellmer", "text-based")
 ) {
   prompt <- tidyprompt(prompt)
   type <- match.arg(type)
@@ -61,7 +73,7 @@ answer_using_tools <- function(
       paste0(
         "{.strong `answer_using_tools()`}:\n",
         "* Automatically determining type based on 'llm_provider$api_type';\n",
-        "this does not consider model compatability\n",
+        "(or 'llm_provider$tool_type' if set); this does not consider model compatability\n",
         "* Manually set argument 'type' if errors occur ",
         "(\"text-based\" always works)\n",
         "* Use `options(tidyprompt.warn.auto.tools = FALSE)` to suppress this warning"
@@ -69,58 +81,107 @@ answer_using_tools <- function(
     )
   }
 
-  # Check if tools is single function, if so, convert to list
-  if (length(tools) == 1 && is.function(tools)) {
-    name <- deparse(substitute(tools))
-    tools <- list(tools)
-    names(tools) <- name
+  # -- Accept function, ellmer ToolDef, or list of either/mixed ---------------
+  is_toolish <- function(x) is.function(x) || is_ellmer_tool(x)
+
+  sanitize_name <- function(x) {
+    x <- as.character(x)
+    x <- gsub("[^A-Za-z0-9_]", "_", x)
+    if (!nzchar(x)) x <- "tool"
+    x
+  }
+
+  # Normalize to a named list while trying to keep stable names
+  if (is_toolish(tools)) {
+    nm <- sanitize_name(deparse(substitute(tools)))
+    tools <- rlang::set_names(list(tools), nm)
   } else {
     stopifnot(
       is.list(tools),
       length(tools) > 0,
-      all(sapply(tools, is.function))
+      all(vapply(tools, is_toolish, logical(1)))
     )
-    # Convert tools to named list
-    tool_names <- sapply(substitute(tools)[-1], deparse)
-    # Ensure the names of tools match the captured names
-    names(tools) <- tool_names
+    if (is.null(names(tools)) || any(!nzchar(names(tools)))) {
+      # Try to derive names from the call; else fallback
+      call_elems <- tryCatch(
+        as.list(substitute(tools))[-1],
+        error = function(e) NULL
+      )
+      if (!is.null(call_elems) && length(call_elems) == length(tools)) {
+        names(tools) <- vapply(
+          call_elems,
+          function(e) sanitize_name(deparse(e)),
+          character(1)
+        )
+      } else {
+        names(tools) <- paste0("tool", seq_along(tools))
+      }
+    } else {
+      names(tools) <- sanitize_name(names(tools))
+    }
+  }
+
+  # -- Build dual representations ---------------------------------------------
+  # For each supplied tool, we want:
+  # - a tidyprompt-style function (with docs) for text-based/OpenAI/Ollama paths
+  # - an ellmer ToolDef for native ellmer path
+  tp_tools <- list()
+  ellmer_tools <- list()
+  for (nm in names(tools)) {
+    dual <- normalize_tool_dual(tools[[nm]])
+    if (!is.null(dual$tidyprompt_tool)) tp_tools[[nm]] <- dual$tidyprompt_tool
+    if (!is.null(dual$ellmer_tool)) ellmer_tools[[nm]] <- dual$ellmer_tool
   }
 
   determine_type <- function(llm_provider = NULL) {
     if (type != "auto") return(type)
+    valid_types <- c(
+      "text-based",
+      "openai",
+      "ollama",
+      "ellmer"
+    )
+    provider_json_type <- llm_provider[["tool_type"]]
+    if (
+      !is.null(provider_json_type) &&
+        !identical(provider_json_type, "auto") &&
+        provider_json_type %in% valid_types
+    ) {
+      return(provider_json_type)
+    }
     if (isTRUE(llm_provider$api_type == "openai")) return("openai")
     if (isTRUE(llm_provider$api_type == "ollama")) return("ollama")
-    return("text-based")
+    if (isTRUE(llm_provider$api_type == "ellmer")) return("ellmer")
+    "text-based"
   }
 
   parameter_fn <- function(llm_provider) {
-    type <- determine_type(llm_provider)
+    t <- determine_type(llm_provider)
     parameters <- list()
 
-    if (type %in% c("openai", "ollama")) {
+    if (t %in% c("openai", "ollama")) {
       tools_openai <- list()
-
-      for (tool_name in names(tools)) {
-        tool_openai <- list()
-        tool <- tools[[tool_name]]
+      for (tool_name in names(tp_tools)) {
+        tool <- tp_tools[[tool_name]]
         docs <- tools_get_docs(tool, tool_name)
 
+        tool_openai <- list()
         tool_openai$type <- "function"
         tool_openai[["function"]]$name <- tool_name
         if (!is.null(docs$description)) {
           tool_openai[["function"]]$description <- docs$description
         }
-        tool_openai[["function"]]$parameters <-
-          tools_docs_to_r_json_schema(docs)
+        tool_openai[["function"]]$parameters <- tools_docs_to_r_json_schema(
+          docs
+        )
         tool_openai[["function"]]$strict <- TRUE
 
         tools_openai[[length(tools_openai) + 1]] <- tool_openai
       }
-
       parameters$tools <- tools_openai
     }
 
-    if (type == "ollama" & isTRUE(llm_provider$parameters$stream)) {
+    if (t == "ollama" & isTRUE(llm_provider$parameters$stream)) {
       cli::cli_alert_warning(
         paste0(
           "{.strong `answer_using_tools()`}:\n",
@@ -131,140 +192,128 @@ answer_using_tools <- function(
       parameters$stream <- FALSE
     }
 
-    return(parameters)
-  }
-
-  handler_fn <- function(response, llm_provider) {
-    type <- determine_type(llm_provider)
-    if (!type %in% c("openai", "ollama")) return(response)
-
-    while (TRUE) {
-      # For as long as there are tool calls
-      body <- response$http$response$body
-
-      if ("tool_calls" %in% names(body)) {
-        # For streaming
-        # As handled by [append_or_update_tool_calls()]
-        tool_calls <- body$tool_calls
-      } else {
-        # Non-streaming; check body for tool calls
-        body <- tryCatch(
-          body |> rawToChar() |> jsonlite::fromJSON(),
-          error = function(e) {
-            NULL
-          }
-        )
-
-        if (is.null(body)) break
-
-        tool_calls <- list()
-
-        if (type == "openai") {
-          if (length(body$choices$message$tool_calls) == 0) break
-
-          for (tool_call in body$choices$message$tool_calls) {
-            tool_calls[[length(tool_calls) + 1]] <- list(
-              id = tool_call$id,
-              type = tool_call$type,
-              "function" = as.list(tool_call[["function"]])
-            )
-          }
-        }
-
-        if (type == "ollama") {
-          if (length(body$message$tool_calls) == 0) break
-
-          for (i in seq_along(body$message$tool_calls)) {
-            tool_call <- body$message$tool_calls[i]
-            tool_calls[[length(tool_calls) + 1]] <- list(
-              "function" = as.list(tool_call[["function"]])
-            )
-          }
+    if (t == "ellmer") {
+      # Ensure we have ToolDefs for every tool; convert if needed
+      missing_defs <- setdiff(names(tp_tools), names(ellmer_tools))
+      if (length(missing_defs)) {
+        for (nm in missing_defs) {
+          # convert tidyprompt tool -> ellmer ToolDef
+          ell_tool <- tryCatch(
+            tidyprompt_tool_to_ellmer(tp_tools[[nm]]),
+            error = function(e) NULL
+          )
+          if (!is.null(ell_tool)) ellmer_tools[[nm]] <- ell_tool
         }
       }
 
+      # Pass ToolDefs for provider to register natively.
+      # NOTE for provider wrapper:
+      #   if (!is.null(params$.ellmer_tools)) {
+      #     for (td in params$.ellmer_tools) chat$register_tool(td)
+      #   }
+      if (length(ellmer_tools)) {
+        parameters$.ellmer_tools <- unname(ellmer_tools)
+      }
+    }
+
+    parameters
+  }
+
+  handler_fn <- function(response, llm_provider) {
+    t <- determine_type(llm_provider)
+    # Native tool handling is done by the provider/model itself
+    if (t == "ellmer") return(response)
+    if (!(t %in% c("openai", "ollama"))) return(response)
+
+    repeat {
+      # 1) Normalize body
+      body <- .parse_http_body(response$http$response$body)
+      if (is.null(body)) break
+
+      # 2) Extract tool calls for this provider type
+      tool_calls <- .get_tool_calls(body, t)
       if (length(tool_calls) == 0) break
 
-      # Add tool calls to messages (to be sent back to API)
+      # 3) Rebuild messages from completed chat history, then add assistant tool_calls
       chat_history <- response$completed
-      messages = lapply(seq_len(nrow(chat_history)), function(i) {
-        list(role = chat_history$role[i], content = chat_history$content[i])
-      })
-      if (type == "openai")
+      messages <- .messages_from_history(chat_history)
+
+      if (t == "openai") {
         messages[[length(messages) + 1]] <- list(
           role = "assistant",
           tool_calls = tool_calls
         )
-      if (type == "ollama")
+      } else if (t == "ollama") {
+        # Keep original Ollama structure if available
         messages[[length(messages) + 1]] <- list(
           role = "assistant",
           tool_calls = body$message$tool_calls
         )
-
-      # For every tool call, call the function and add the result to the messages
-      for (tool_call in tool_calls) {
-        tool_name <- tool_call[["function"]]$name
-        tool <- tools[[tool_name]]
-
-        if (type == "ollama") {
-          arguments <- tool_call[["function"]]$arguments |> as.list()
-        } else {
-          arguments <- tool_call[["function"]]$arguments |> jsonlite::fromJSON()
-        }
-
-        # Also add the tool call to the chat history
-        chat_history <- chat_history |>
-          dplyr::bind_rows(
-            data.frame(
-              role = "assistant",
-              content = paste0(
-                "Calling function '",
-                tool_name,
-                "' with arguments:\n",
-                jsonlite::toJSON(arguments, auto_unbox = TRUE, pretty = TRUE)
-              )
-            )
-          )
-
-        # Execute tool function
-        result <- tryCatch(
-          {
-            do.call(tool, arguments)
-          },
-          error = function(e) {
-            glue::glue("Error: {e$message}")
-          }
-        )
-
-        # Ensure result is proper string
-        if (length(result) > 0) result <- paste(result, collapse = ", ")
-        result <- as.character(result)
-
-        # Add tool result to messages
-        message_addition <- list(role = "tool", content = result)
-        if (type == "openai") message_addition$tool_call_id <- tool_call$id
-        messages[[length(messages) + 1]] <- message_addition
-
-        # Also add tool result to chat history
-        chat_history <- chat_history |>
-          dplyr::bind_rows(
-            data.frame(
-              role = "tool",
-              content = paste0("Result:\n", result),
-              tool_call_id = ifelse(is.null(tool_call$id), NA, tool_call$id),
-              tool_result = TRUE
-            )
-          )
-
-        # Print directly if streaming
-        #   (if verbose & no stream, it will be printed in llm_provider$complete_chat())
-        if (isTRUE(llm_provider$parameters$stream))
-          cat(paste0("Result:\n", result, "\n"))
       }
 
-      # Finally new request with tool results
+      # 4) Iterate over tool calls: parse args, call tool, record result
+      for (tool_call in tool_calls) {
+        tool_name <- tool_call[["function"]]$name
+        tool <- tp_tools[[tool_name]]
+
+        # Parse arguments robustly
+        arguments <- .parse_arguments(tool_call, t)
+
+        # Record the call in chat history
+        pretty_args <- tryCatch(
+          jsonlite::toJSON(arguments, auto_unbox = FALSE, pretty = TRUE),
+          error = function(e) "null"
+        )
+        chat_history <- .append_history(
+          chat_history,
+          role = "assistant",
+          content = paste0(
+            "~>> Calling function '",
+            tool_name,
+            "' with arguments:\n",
+            pretty_args
+          ),
+          tool_call_id = tool_call$id,
+          tool_call = TRUE,
+          tool_result = FALSE
+        )
+        .maybe_message(
+          dplyr::slice_tail(chat_history, n = 1)$content,
+          llm_provider
+        )
+
+        # Call the tool safely
+        result <- tryCatch(
+          do.call(tool, arguments),
+          error = function(e) glue::glue("Error: {e$message}")
+        )
+
+        # Normalize result to character
+        if (length(result) > 0) result <- paste(result, collapse = ", ")
+        if (length(result) == 0) result <- ""
+        result <- as.character(result)
+
+        # Add tool result to next request messages
+        message_addition <- list(role = "tool", content = result)
+        if (t == "openai") message_addition$tool_call_id <- tool_call$id
+        messages[[length(messages) + 1]] <- message_addition
+
+        # Record result in chat history
+        chat_history <- .append_history(
+          chat_history,
+          role = "tool",
+          content = paste0("~>> Result:\n", result),
+          tool_call_id = tool_call$id,
+          tool_call = FALSE,
+          tool_result = TRUE
+        )
+        .maybe_message(paste0("~>> Result:\n", result), llm_provider)
+      }
+
+      # 5) Send follow-up request including tool outputs, keep loop if model asks again
       new_request <- response$http$request
       new_request$body$data$messages <- messages
+
       response <- request_llm_provider(
         chat_history,
         new_request,
@@ -274,12 +323,12 @@ answer_using_tools <- function(
       )
     }
 
-    return(response)
+    response
   }
 
   modify_fn <- function(original_prompt_text, llm_provider) {
-    type <- determine_type(llm_provider)
-    if (type %in% c("openai", "ollama")) return(original_prompt_text)
+    t <- determine_type(llm_provider)
+    if (t %in% c("openai", "ollama", "ellmer")) return(original_prompt_text)
 
     new_prompt <- glue::glue(
       "{original_prompt_text}\n\n",
@@ -304,19 +353,12 @@ answer_using_tools <- function(
       "The following functions are available:"
     )
 
-    for (tool_name in names(tools)) {
-      tool <- tools[[tool_name]]
+    for (tool_name in names(tp_tools)) {
+      tool <- tp_tools[[tool_name]]
       docs <- tools_get_docs(tool, tool_name)
 
-      with_arguments <- TRUE
-      if (type == "openai") with_arguments <- FALSE
-
-      tool_llm_text <- tools_docs_to_text(docs, with_arguments)
-
-      new_prompt <- glue::glue(
-        "{new_prompt}\n\n{tool_llm_text}",
-        .trim = FALSE
-      )
+      tool_llm_text <- tools_docs_to_text(docs, with_arguments = TRUE)
+      new_prompt <- glue::glue("{new_prompt}\n\n{tool_llm_text}", .trim = FALSE)
     }
 
     new_prompt <- glue::glue(
@@ -326,20 +368,20 @@ answer_using_tools <- function(
       .trim = FALSE
     )
 
-    return(new_prompt)
+    new_prompt
   }
 
   extraction_fn <- function(llm_response, llm_provider) {
-    type <- determine_type(llm_provider)
-    if (type %in% c("openai", "ollama")) return(llm_response)
+    t <- determine_type(llm_provider)
+    if (t %in% c("openai", "ollama", "ellmer")) return(llm_response)
 
     jsons <- extraction_fn_json(llm_response)
 
     fn_results <- lapply(jsons, function(json) {
       if (is.null(json[["function"]])) return(NULL)
 
-      if (json[["function"]] %in% names(tools)) {
-        tool_function <- tools[[json[["function"]]]]
+      if (json[["function"]] %in% names(tp_tools)) {
+        tool_function <- tp_tools[[json[["function"]]]]
         arguments <- json[["arguments"]]
 
         result <- tryCatch(
@@ -352,40 +394,31 @@ answer_using_tools <- function(
         )
 
         if (length(result) > 0) result <- paste(result, collapse = ", ")
+        if (length(result) == 0) result <- ""
 
-        # Create some context around the result
         string_of_named_arguments <-
           paste(names(arguments), arguments, sep = " = ") |>
-            paste(collapse = ", ")
+          paste(collapse = ", ")
 
-        result_string <- glue::glue(
-          "function called: {json[[\"function\"]]}
-          arguments used: {string_of_named_arguments}
-          result: {result}"
+        glue::glue(
+          "function called: {json[[\"function\"]]}\n",
+          "arguments used: {string_of_named_arguments}\n",
+          "result: {result}"
         )
-
-        return(result_string)
       } else {
-        return(
-          glue::glue("Error: Function '{json[[\"function\"]]}' not found.")
-        )
+        glue::glue("Error: Function '{json[[\"function\"]]}' not found.")
       }
     })
-    fn_results <- fn_results[!sapply(fn_results, is.null)]
+    fn_results <- fn_results[!vapply(fn_results, is.null, logical(1))]
 
     if (length(fn_results) == 0) return(llm_response)
 
-    return(
-      llm_feedback(
-        paste(fn_results, collapse = "\n\n"),
-        tool_result = TRUE
-      )
-    )
+    llm_feedback(paste(fn_results, collapse = "\n\n"), tool_result = TRUE)
   }
 
-  # Add environment with tool functions as an attribute to the extraction function
+  # Attach the (tidyprompt) tool functions for text-based/OpenAI/Ollama paths
   environment_with_tools <- new.env()
-  environment_with_tools$tools <- tools
+  environment_with_tools$tools <- tp_tools
   attr(extraction_fn, "environment") <- environment_with_tools
   attr(handler_fn, "environment") <- environment_with_tools
 
@@ -401,6 +434,173 @@ answer_using_tools <- function(
   )
 }
 
+
+# Helpers for handler_fn in `answer_using_tools()` ------------------------
+
+# These functions are used by the handler_fn in `answer_using_tools()`,
+#  to parse tool calls from LLM responses of various API providers,
+#  normalize them, call the relevant R functions, and provide the results
+#  back to the LLM
+# Used for text-based, openai, and ollama types; when using an 'ellmer'
+#   LLM provider, tool calling is handled natively by 'ellmer'
+
+.parse_http_body <- function(resp_body) {
+  # If it's already a list with tool_calls (some wrappers pre-normalize), return it
+  if (is.list(resp_body) && ("tool_calls" %in% names(resp_body))) {
+    return(resp_body)
+  }
+  # Try to decode raw -> JSON -> list
+  if (is.raw(resp_body)) {
+    return(tryCatch(
+      jsonlite::fromJSON(rawToChar(resp_body)),
+      error = function(e) NULL
+    ))
+  }
+  # If it's character JSON, try to parse
+  if (is.character(resp_body) && length(resp_body) == 1) {
+    return(tryCatch(
+      jsonlite::fromJSON(resp_body),
+      error = function(e) NULL
+    ))
+  }
+  # If it's already a list but without tool_calls, just return it
+  if (is.list(resp_body)) return(resp_body)
+  NULL
+}
+
+.normalize_openai_tool_calls <- function(body) {
+  calls <- list()
+
+  # OpenAI Chat Completions-style: body$choices$message$tool_calls
+  if (
+    !is.null(body$choices) &&
+      !is.null(body$choices$message) &&
+      length(body$choices$message$tool_calls) > 0
+  ) {
+    for (tc in body$choices$message$tool_calls) {
+      calls[[length(calls) + 1]] <- list(
+        id = tc$id,
+        type = tc$type,
+        "function" = as.list(tc[["function"]])
+      )
+    }
+    return(calls)
+  }
+
+  # OpenAI Responses API-style: data.frame body$output with type == "function_call"
+  if (
+    !is.null(body$output) &&
+      is.data.frame(body$output) &&
+      nrow(body$output) > 0 &&
+      any(body$output$type == "function_call")
+  ) {
+    fc <- body$output[body$output$type == "function_call", , drop = FALSE]
+    for (i in seq_len(nrow(fc))) {
+      calls[[length(calls) + 1]] <- list(
+        id = fc$id[i],
+        type = fc$type[i],
+        "function" = list(
+          name = fc$name[i],
+          arguments = fc$arguments[i]
+        )
+      )
+    }
+    return(calls)
+  }
+
+  # Nothing found
+  calls
+}
+
+.normalize_ollama_tool_calls <- function(body) {
+  calls <- list()
+  if (
+    !is.null(body$message) &&
+      length(body$message$tool_calls) > 0
+  ) {
+    for (i in seq_along(body$message$tool_calls)) {
+      tc <- body$message$tool_calls[i]
+      calls[[length(calls) + 1]] <- list(
+        "function" = as.list(tc[["function"]])
+      )
+    }
+  }
+  calls
+}
+
+.get_tool_calls <- function(body, t) {
+  # Case where upstream wrapped a tool_calls list directly in body
+  if (!is.null(body$tool_calls) && length(body$tool_calls) > 0) {
+    return(body$tool_calls)
+  }
+  if (t == "openai") return(.normalize_openai_tool_calls(body))
+  if (t == "ollama") return(.normalize_ollama_tool_calls(body))
+  list()
+}
+
+.parse_arguments <- function(tool_call, t) {
+  if (t == "ollama") {
+    return(as.list(tool_call[["function"]]$arguments))
+  }
+  # OpenAI: try "arguments" then "args", both JSON
+  args <- NULL
+  try(
+    {
+      args <- jsonlite::fromJSON(tool_call[["function"]]$arguments)
+    },
+    silent = TRUE
+  )
+  if (is.null(args)) {
+    try(
+      {
+        args <- jsonlite::fromJSON(tool_call[["function"]]$args)
+      },
+      silent = TRUE
+    )
+  }
+  args
+}
+
+.append_history <- function(
+  history,
+  role,
+  content,
+  tool_call_id = NA,
+  tool_call = FALSE,
+  tool_result = FALSE
+) {
+  dplyr::bind_rows(
+    history,
+    data.frame(
+      role = role,
+      content = content,
+      tool_call = tool_call,
+      tool_call_id = ifelse(is.null(tool_call_id), NA, tool_call_id),
+      tool_result = tool_result
+    )
+  )
+}
+
+.maybe_message <- function(text, llm_provider) {
+  if (isTRUE(llm_provider$parameters$stream) || isTRUE(llm_provider$verbose)) {
+    message(text)
+    message("")
+  }
+}
+
+.messages_from_history <- function(history_df) {
+  lapply(seq_len(nrow(history_df)), function(i) {
+    list(role = history_df$role[i], content = history_df$content[i])
+  })
+}
+
+
+# Add documentation to functions ------------------------------------------
+
+# These functions are used to extract & add documentation to/from R functions,
+#   so that LLMs can be provided with information about the functions
+# They are sort of superseded by the 'ellmer' integration and `ellmer::tool()`
+
 #' Add tidyprompt function documentation to a function
 #'
 #' @description This function adds documentation to a custom function. This documentation
@@ -531,10 +731,8 @@ tools_get_docs <- function(func, name = NULL) {
       is.list(docs$arguments),
       !is.null(names(docs$arguments)),
       is.null(docs$return$description) ||
-        (
-          is.character(docs$return$description) &
-            length(docs$return$description) == 1
-        )
+        (is.character(docs$return$description) &
+          length(docs$return$description) == 1)
     )
   } else {
     docs <- tools_generate_docs(name, func)
@@ -1007,6 +1205,23 @@ tools_docs_to_r_json_schema <- function(
     prop <- list()
     r_type <- arg$type
 
+    normalize_r_type <- function(x) {
+      if (is.list(x) || is.null(x)) return(x)
+      x <- tolower(trimws(as.character(x)))
+      switch(
+        x,
+        "character" = "string",
+        "str" = "string",
+        "bool" = "logical",
+        "boolean" = "logical",
+        "double" = "numeric",
+        "float" = "numeric",
+        "int" = "integer",
+        x
+      )
+    }
+    r_type <- normalize_r_type(arg$type)
+
     if (is.list(r_type)) {
       # Handle named lists (objects)
       prop$type <- "object"
@@ -1022,7 +1237,7 @@ tools_docs_to_r_json_schema <- function(
       }
     } else if (is.null(r_type) || r_type == "unknown") {
       prop$type <- "string"
-    } else if (r_type == "character") {
+    } else if (r_type == "string") {
       prop$type <- "string"
     } else if (r_type == "integer") {
       prop$type <- "integer"
@@ -1106,7 +1321,10 @@ tools_docs_to_r_json_schema <- function(
       cli::cli_alert_warning(
         paste0(
           "{.strong `answer_using_tools()`, `tools_docs_to_r_json_schema()`}:\n",
-          "* Unknown argument type; defaulting to 'string'"
+          "* Unknown argument type ('",
+          r_type,
+          "'); ",
+          "defaulting to 'string'"
         )
       )
       prop$type <- "string"
@@ -1137,7 +1355,7 @@ tools_docs_to_r_json_schema <- function(
   list(
     type = "object",
     properties = properties,
-    required = required_args,
+    required = I(required_args),
     additionalProperties = additional_properties
   )
 }
