@@ -36,16 +36,50 @@ NULL
     api_key = NULL,
 
     #' @field api_type
-    #' The type of API to use (e.g., "openai", "ollama").
+    #' The type of API to use (e.g., "openai", "ollama", "ellmer").
     #'  This is used to determine certain specific behaviors for different APIs,
     #'  for instance, as is done in the [answer_as_json()] function
     api_type = "unspecified",
 
+    #' @field json_type
+    #' The type of JSON mode to use (e.g., 'auto', 'openai', 'ollama', 'ellmer', or 'text-based').
+    #'  Using 'auto' or having this field not set, the api_type field will be used to
+    #'  determine the JSON mode during the [answer_as_json()] function. If this field
+    #'  is set, this will override the api_type field for JSON mode determination.
+    #'  (Note: this determination only happens when the 'type' argument in
+    #'  [answer_as_json()] is also set to 'auto'.)
+    json_type = "auto",
+
+    #' @field tool_type
+    #' The type of tool use mode to use (e.g., 'auto', 'openai', 'ollama', 'ellmer', or 'text-based').
+    #'  Using 'auto' or having this field not set, the api_type field will be used to
+    #'  determine the tool use mode during the [answer_using_tools()] function. If this field
+    #'  is set, this will override the api_type field for tool use mode determination
+    #'  (Note: this determination only happens when the 'type' argument in
+    #'  [answer_using_tools()] is also set to 'auto'.)
+    tool_type = "auto",
+
     #' @field handler_fns
     #' A list of functions that will be called after the completion of a chat.
     #'  See `$add_handler_fn()`
     handler_fns = list(),
 
+    #' @field pre_prompt_wraps
+    #' A list of prompt wraps that will be applied to any prompt evaluated
+    #' by this [llm_provider-class] object, before any prompt-specific
+    #' prompt wraps are applied. See `$add_prompt_wrap()`.
+    #' This can be used to set default behavior for all prompts
+    #' evaluated by this [llm_provider-class] object.
+    pre_prompt_wraps = list(),
+
+    #' @field post_prompt_wraps
+    #' A list of prompt wraps that will be applied to any prompt evaluated
+    #' by this [llm_provider-class] object, after any prompt-specific
+    #' prompt wraps are applied. See `$add_prompt_wrap()`.
+    #' This can be used to set default behavior for all prompts
+    #' evaluated by this [llm_provider-class] object.
+    post_prompt_wraps = list(),
+
     #' @description
     #' Create a new [llm_provider-class] object
     #'
@@ -191,6 +225,13 @@ NULL
       environment(private$complete_chat_function) <- environment()
       response <- private$complete_chat_function(chat_history)
 
+      # If this is an ellmer provider, sync the chat object
+      #   This ensures handler_fn can access the current state of the chat
+      #   object (e.g., to monitor cost and cancel if over budget)
+      if (!is.null(response$ellmer_chat) && "ellmer_chat" %in% names(self)) {
+        self$ellmer_chat <- response$ellmer_chat
+      }
+
       # Filter content with empty string ("") (Ollama tool call)
       response$completed <- response$completed[
         response$completed$content != "",
@@ -227,14 +268,31 @@ NULL
       # Print difference between chat_history and completed
       if (
         self$verbose &&
-        (is.null(self$parameters$stream) || !self$parameters$stream)
+          (is.null(self$parameters$stream) || !self$parameters$stream)
       ) {
         chat_history_new <- response$completed[
           (nrow(chat_history) + 1):nrow(response$completed),
         ]
 
-        for (i in seq_len(nrow(chat_history_new))) {
-          message(chat_history_new$content[i])
+        # Filter out rows with 'tool_result' == TRUE
+        # That's already being printed in the handler function of
+        #   `answer_using_tools()`
+        if (
+          "tool_result" %in%
+            names(chat_history_new) &
+            "tool_call" %in% names(chat_history_new)
+        ) {
+          chat_history_new_print <- chat_history_new |>
+            dplyr::filter(
+              (is.na(tool_result) | tool_result == FALSE),
+              (is.na(tool_call) | tool_call == FALSE)
+            )
+        } else {
+          chat_history_new_print <- chat_history_new
+        }
+
+        for (i in seq_len(nrow(chat_history_new_print))) {
+          message(chat_history_new_print$content[i])
         }
       }
 
@@ -289,6 +347,110 @@ NULL
       stopifnot(is.list(handler_fns))
       self$handler_fns <- handler_fns
       return(self)
+    },
+
+    #' @description
+    #' Add a provider-level prompt wrap template to be applied to all prompts.
+    #' @param prompt_wrap A list created by [provider_prompt_wrap()]
+    #' @param position One of "pre" or "post" (applied before/after prompt-specific wraps)
+    add_prompt_wrap = function(prompt_wrap, position = c("pre", "post")) {
+      position <- match.arg(position)
+      stopifnot(is.list(prompt_wrap))
+
+      # Normalize fields so downstream code can rely on them
+      needed <- c(
+        "type",
+        "modify_fn",
+        "extraction_fn",
+        "validation_fn",
+        "handler_fn",
+        "parameter_fn",
+        "name"
+      )
+      missing <- setdiff(needed, names(prompt_wrap))
+      for (nm in missing) prompt_wrap[[nm]] <- NULL
+
+      class(prompt_wrap) <- unique(c(
+        "provider_prompt_wrap",
+        class(prompt_wrap)
+      ))
+
+      if (identical(position, "pre")) {
+        self$pre_prompt_wraps <- c(self$pre_prompt_wraps, list(prompt_wrap))
+      } else {
+        self$post_prompt_wraps <- c(self$post_prompt_wraps, list(prompt_wrap))
+      }
+      invisible(self)
+    },
+
+    #' @description
+    #' Apply all provider-level wraps to a prompt (character or tidyprompt)
+    #' and return a tidyprompt with wraps attached.
+    #' This is typically called inside `send_prompt()` before evaluation of
+    #' the prompt.
+    #' @param prompt A string, a chat history, a list containing
+    #' a chat history under key '$chat_history', or a [tidyprompt-class] object
+    apply_prompt_wraps = function(prompt) {
+      if (!inherits(prompt, "Tidyprompt")) {
+        prompt <- tidyprompt(prompt)
+      }
+
+      # Fresh Tidyprompt with same base/system/chat_history but no wraps yet
+      new_prompt <- `tidyprompt-class`$new(prompt$base_prompt)
+      new_prompt$system_prompt <- prompt$system_prompt
+      # Preserve chat history without constructing prompt text
+      new_prompt$.__enclos_env__$private$chat_history <-
+        prompt$.__enclos_env__$private$chat_history
+
+      # 1) Provider pre wraps
+      if (length(self$pre_prompt_wraps)) {
+        for (pw in self$pre_prompt_wraps) {
+          new_prompt <- prompt_wrap_internal(
+            new_prompt,
+            modify_fn = pw$modify_fn,
+            extraction_fn = pw$extraction_fn,
+            validation_fn = pw$validation_fn,
+            handler_fn = pw$handler_fn,
+            parameter_fn = pw$parameter_fn,
+            type = pw$type,
+            name = pw$name
+          )
+        }
+      }
+
+      # 2) Existing prompt wraps in their original order
+      if (length(prompt$get_prompt_wraps(order = "default"))) {
+        for (pw in prompt$get_prompt_wraps(order = "default")) {
+          new_prompt <- prompt_wrap_internal(
+            new_prompt,
+            modify_fn = pw$modify_fn,
+            extraction_fn = pw$extraction_fn,
+            validation_fn = pw$validation_fn,
+            handler_fn = pw$handler_fn,
+            parameter_fn = pw$parameter_fn,
+            type = pw$type,
+            name = pw$name
+          )
+        }
+      }
+
+      # 3) Provider post wraps
+      if (length(self$post_prompt_wraps)) {
+        for (pw in self$post_prompt_wraps) {
+          new_prompt <- prompt_wrap_internal(
+            new_prompt,
+            modify_fn = pw$modify_fn,
+            extraction_fn = pw$extraction_fn,
+            validation_fn = pw$validation_fn,
+            handler_fn = pw$handler_fn,
+            parameter_fn = pw$parameter_fn,
+            type = pw$type,
+            name = pw$name
+          )
+        }
+      }
+
+      new_prompt
     }
   ),
   private = list(
