@@ -109,6 +109,7 @@ llm_provider_ollama <- function(
 #' Create a new OpenAI LLM provider
 #'
 #' This function creates a new [llm_provider-class] object that interacts with the Open AI API.
+#' Supports both the Chat Completions API (/v1/chat/completions) and the Responses API (/v1/responses).
 #'
 #' @param parameters A named list of parameters. Currently the following parameters are required:
 #'    - model: The name of the model to use
@@ -123,7 +124,9 @@ llm_provider_ollama <- function(
 #' @param verbose A logical indicating whether the interaction with the LLM provider
 #' should be printed to the console. Default is TRUE.
 #' @param url The URL to the OpenAI API endpoint for chat completion
-#' (typically: "https://api.openai.com/v1/chat/completions")
+#' (typically: "https://api.openai.com/v1/chat/completions"
+#' or "https://api.openai.com/v1/responses"; both the Chat Completions API and
+#' the Responses API are supported; the Responses API is more modern)
 #' @param api_key The API key to use for authentication with the OpenAI API
 #'
 #' @return A new [llm_provider-class] object for use of the OpenAI API
@@ -138,7 +141,7 @@ llm_provider_openai <- function(
     stream = getOption("tidyprompt.stream", TRUE)
   ),
   verbose = getOption("tidyprompt.verbose", TRUE),
-  url = "https://api.openai.com/v1/chat/completions",
+  url = "https://api.openai.com/v1/responses",
   api_key = Sys.getenv("OPENAI_API_KEY")
 ) {
   complete_chat <- function(chat_history) {
@@ -314,8 +317,16 @@ llm_provider_xai <- function(
 
 #' Create a new Google Gemini LLM provider
 #'
+#' @description
+#' `r lifecycle::badge("superseded")`
 #' This function creates a new [llm_provider-class] object that interacts with the Google Gemini API.
-#' Streaming is not yet supported in this implementation.
+#'
+#' @details
+#' Streaming is not yet supported in this implementation. Native functions
+#' like structured output and tool calling are also not supported in this implemetation.
+#' This may however be achieved through creating a [llm_provider_ellmer()] object
+#' with as input a `ellmer::chat_google_gemini()` object. Therefore, this function
+#' is now superseded by `llm_provider_ellmer(ellmer::chat_google_gemini())`.
 #'
 #' @param parameters A named list of parameters. Currently the following parameters are required:
 #'    - model: The name of the model to use (see: https://ai.google.dev/gemini-api/docs/models/gemini)
@@ -696,6 +707,12 @@ llm_provider_fake <- function(verbose = getOption("tidyprompt.verbose", TRUE)) {
 #' (and not in the `$parameters` list). `$get_chat()` and `$set_chat()` may be used
 #' to manipulate the chat object.
 #'
+#' A special parameter `$.ellmer_structured_type` may be set in the `$parameters` list;
+#' this parameter is used to specify a structured output format. This should be a 'ellmer'
+#' structured type (e.g., `ellmer::type_object`; see https://ellmer.tidyverse.org/articles/structured-data.html).
+#' `answer_as_json()` sets this parameter to obtain structured output
+#' (it is not recommended to set this parameter manually, but it is possible).
+#'
 #' @param chat An `ellmer::chat()` object (e.g., `ellmer::chat_openai()`)
 #' @param verbose A logical indicating whether the interaction with the [llm_provider-class]
 #' should be printed to the console. Default is TRUE
@@ -708,26 +725,28 @@ llm_provider_fake <- function(verbose = getOption("tidyprompt.verbose", TRUE)) {
 #'
 #' @family llm_provider
 llm_provider_ellmer <- function(
-    chat,
-    verbose = getOption("tidyprompt.verbose", TRUE)
+  chat,
+  verbose = getOption("tidyprompt.verbose", TRUE)
 ) {
   if (missing(chat) || is.null(chat)) {
     stop("`chat` must be an ellmer chat object (e.g., ellmer::chat_openai()).")
   }
   if (!is.environment(chat) || !is.function(chat$chat)) {
-    stop("`chat` doesn't look like an ellmer chat object (no `$chat()` method).")
+    stop(
+      "`chat` doesn't look like an ellmer chat object (no `$chat()` method)."
+    )
   }
 
   complete_chat <- function(chat_history) {
     private$sync_model()
     ch <- self$ellmer_chat
-    ch <- ch$clone()
+    params <- self$parameters
 
     if (!all(c("role", "content") %in% names(chat_history))) {
       stop("`chat_history` must have columns `role` and `content`.")
     }
 
-    # All but last message become seeded turns
+    # Seed prior turns
     hist <- if (nrow(chat_history) > 1) {
       chat_history[seq_len(nrow(chat_history) - 1), , drop = FALSE]
     } else {
@@ -735,28 +754,45 @@ llm_provider_ellmer <- function(
     }
 
     to_turn <- function(role, text) {
-      ellmer::Turn(
-        role = role,
-        contents = list(ellmer::ContentText(text))
-      )
+      ellmer::Turn(role = role, contents = list(ellmer::ContentText(text)))
     }
-    turns <- if (nrow(hist)) {
-      lapply(seq_len(nrow(hist)), function(i) to_turn(hist$role[i], hist$content[i]))
+    if (nrow(hist)) {
+      ch <- ch$set_turns(lapply(seq_len(nrow(hist)), function(i) {
+        to_turn(hist$role[i], hist$content[i])
+      }))
     } else {
-      NULL
+      ch <- ch$set_turns(list())
     }
 
-    ch <- ch$set_turns(turns)
-
-    # Last message is always the prompt
+    # Prompt = last message
     prompt <- chat_history$content[nrow(chat_history)]
-    reply <- ch$chat(prompt)
+
+    # Register tools
+    if (!is.null(params$.ellmer_tools)) {
+      for (td in params$.ellmer_tools) ch$register_tool(td)
+    }
+
+    # Check if we are doing structured output
+    structured_type <- params$.ellmer_structured_type %||% NULL
+    use_structured <- !is.null(structured_type) &&
+      is.function(ch$chat_structured)
+
+    if (use_structured) {
+      reply_struct <- ch$chat_structured(prompt, type = structured_type)
+      # Store a JSON string in the transcript (so downstream plain JSON extractors still work)
+      assistant_text <- jsonlite::toJSON(reply_struct, auto_unbox = TRUE) |>
+        as.character()
+    } else {
+      # Regular chat
+      reply_any <- ch$chat(prompt)
+      assistant_text <- as.character(reply_any)
+    }
 
     completed <- dplyr::bind_rows(
       chat_history,
       data.frame(
         role = "assistant",
-        content = as.character(reply),
+        content = assistant_text,
         stringsAsFactors = FALSE
       )
     )
@@ -764,10 +800,7 @@ llm_provider_ellmer <- function(
     list(
       completed = completed,
       http = list(request = NULL, response = NULL),
-      ellmer = list(
-        chat = ch,
-        turns = tryCatch(ch$get_turns(), error = function(e) NULL)
-      )
+      ellmer_chat = ch
     )
   }
 
@@ -792,10 +825,13 @@ llm_provider_ellmer <- function(
 
       # Keep self$parameters$model in sync with ellmer_chat$get_model()
       sync_model = function() {
-        m <- tryCatch({
-          gm <- self$ellmer_chat$get_model
-          if (is.function(gm)) self$ellmer_chat$get_model() else NULL
-        }, error = function(e) NULL)
+        m <- tryCatch(
+          {
+            gm <- self$ellmer_chat$get_model
+            if (is.function(gm)) self$ellmer_chat$get_model() else NULL
+          },
+          error = function(e) NULL
+        )
         if (!is.null(m)) self$parameters$model <- as.character(m)
         invisible(NULL)
       }
@@ -815,4 +851,3 @@ llm_provider_ellmer <- function(
 
   provider
 }
-
