@@ -54,9 +54,17 @@
 #' last message from the assistant, and all messages from the system are kept),
 #'    \item 'start_time' (the time when the function was called),
 #'    \item 'end_time' (the time when the function ended),
-#'    \item 'duration_seconds' (the duration of the function in seconds), and
+#'    \item 'duration_seconds' (the duration of the function in seconds),
 #'    \item 'http_list' (a list with all HTTP responses made during the interactions;
-#'    as returned by `llm_provider$complete_chat()`).
+#'    as returned by `llm_provider$complete_chat()`),
+#'    \item 'ellmer_chat' (if [llm_provider_ellmer()] was used, this will be
+#'    the updated 'ellmer' chat object, containing for instance
+#'    the turns and possible tool calls. (As this function
+#'    uses a clone of the provided LLM provider, the 'ellmer' chat object in the
+#'    LLM provider will not be updated; via this way, you can
+#'    then still get an updated 'ellmer' chat object. Note that turns in the 'ellmer'
+#'    chat object may not contain the full chat history when `clean_chat_history = TRUE`
+#'    was used.)
 #'  }
 #' }
 #'
@@ -72,7 +80,7 @@ send_prompt <- function(
   prompt,
   llm_provider = llm_provider_ollama(),
   max_interactions = 10,
-  clean_chat_history = TRUE,
+  clean_chat_history = FALSE,
   verbose = NULL,
   stream = NULL,
   return_mode = c("only_response", "full")
@@ -91,6 +99,11 @@ send_prompt <- function(
     is.null(stream) | is.logical(stream)
   )
 
+  # Add provider-specific pre/post prompt wraps
+  if (isTRUE(is.function(llm_provider$apply_prompt_wraps))) {
+    prompt <- llm_provider$apply_prompt_wraps(prompt)
+  }
+
   # Verify and configure llm_provider
   llm_provider <- llm_provider$clone()
   if (!is.null(verbose)) llm_provider$verbose <- verbose
@@ -115,10 +128,13 @@ send_prompt <- function(
   # Initialize variables which keep track of the process
   if (return_mode == "full") start_time <- Sys.time()
   http <- list(requests = list(), responses = list())
+  # Object which keeps ellmer chat object, turns, structured output;
+  #   when using an ellmer LLM provider:
+  ellmer_chat <- NULL
 
   ## 2 Chat history, send_chat, handler_fns
 
-  chat_history <- prompt$get_chat_history()
+  chat_history <- prompt$get_chat_history(llm_provider)
 
   # Internal function to send chat messages
   send_chat <- function(
@@ -156,6 +172,10 @@ send_prompt <- function(
     for (http_request in response$http$request)
       http$requests[[length(http$requests) + 1]] <<- http_request
 
+    if (!is.null(response$ellmer_chat)) {
+      ellmer_chat <<- response$ellmer_chat
+    }
+
     utils::tail(chat_history$content, 1)
   }
 
@@ -340,9 +360,14 @@ send_prompt <- function(
     return_list$response <- response
     return_list$interactions <- interactions
     return_list$chat_history <- chat_history
-    return_list$chat_history_clean <- clean_chat_history(chat_history)
+
+    if (clean_chat_history) {
+      return_list$chat_history_clean <- clean_chat_history(chat_history)
+    }
+
     return_list$start_time <- start_time
     return_list$end_time <- Sys.time()
+
     return_list$duration_seconds <-
       as.numeric(
         difftime(
@@ -351,14 +376,14 @@ send_prompt <- function(
           units = "secs"
         )
       )
+
     return_list$http <- http
+    return_list$ellmer_chat <- ellmer_chat
 
     return(return_list)
   }
 }
 
-# Create internal function which cleans the chat_history
-
 #' Create chat history dataframe
 #'
 #' @description Internal function used by [send_prompt()] to create a chat history
