@@ -283,18 +283,35 @@ NULL
     #'
     #' @param llm_provider Optional [llm_provider-class] object.
     #' This may sometimes affect the prompt text construction
+    #' @param apply_provider_prompt_wraps Logical. Whether to apply
+    #' provider-specific pre/post prompt wraps when constructing the prompt text
     #'
     #' @return A string representing the constructed prompt text
-    construct_prompt_text = function(llm_provider = NULL) {
+    construct_prompt_text = function(
+      llm_provider = NULL,
+      apply_provider_prompt_wraps = FALSE
+    ) {
       private$validate_tidyprompt()
       prompt_text <- self$base_prompt
-      wraps <- self$get_prompt_wraps(order = "modification")
+
+      # Add provider-specific pre/post prompt wraps
+      if (
+        isTRUE(is.function(llm_provider$apply_prompt_wraps)) &
+          isTRUE(apply_provider_prompt_wraps)
+      ) {
+        modified_self <- llm_provider$apply_prompt_wraps(self)
+      } else {
+        modified_self <- self
+      }
+
+      wraps <- modified_self$get_prompt_wraps(order = "modification")
 
       for (wrap in wraps) {
         if (!is.null(wrap$modify_fn)) {
           prompt_text <- wrap$modify_fn(prompt_text, llm_provider)
         }
       }
+
       prompt_text
     },
 
@@ -343,8 +360,11 @@ NULL
     #' with the system prompt as the first message with role 'system' and the
     #' the base prompt as the last message with role 'user'.
     #'
+    #' @param llm_provider An optional [llm_provider-class] object.
+    #' This may sometimes affect the prompt text construction
+    #'
     #' @return A dataframe containing the chat history
-    get_chat_history = function() {
+    get_chat_history = function(llm_provider = NULL) {
       chat_history_construction <- c(
         role = "system",
         content = self$system_prompt
@@ -355,7 +375,7 @@ NULL
         dplyr::bind_rows(
           c(
             role = "user",
-            content = self$construct_prompt_text()
+            content = self$construct_prompt_text(llm_provider)
           )
         )
 
